module.exports = (function () {
    var type = require("type-detect");
    var isNil = require("is-nil");
    var clone = require("clone");
    var associative = require("associative");
    var ArrayX = (function () {
        function ArrayX() {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var Arr = new Array(0);
            args.forEach(function (arg) {
                Arr.push(arg);
            });
            Arr["forAll"] = function (func) {
                return this.forEach(func);
            };
            Arr["distinct"] = function () {
                var arr = [];
                for (var i = 0; i < this.length; i++) {
                    if (arr.indexOf(this[i]) < 0)
                        arr.push(this[i]);
                }
                return new ArrayX(arr);
            };
            Arr["accumulate"] = function (initial, accum) {
                var val = initial; // initial parameter is required
                for (var i = 0; i < this.length; i++)
                    val = accum ? accum(accum, this[i]) : (val + this[i]);
                return val;
            };
            Arr["union"] = function (array) {
                var arr = [];
                for (var i = 0; i < this.length; i++)
                    arr.push(this[i]);
                if (type(array) === "array")
                    for (var k = 0; k < array.length; k++)
                        arr.push(array[k]);
                return new ArrayX(arr);
            };
            Arr["select"] = function (func) {
                var arr = [];
                for (var i = 0; i < this.length; i++)
                    arr.push(func ? func(this[i], i) : this[i]);
                return new ArrayX(arr);
            };
            Arr["where"] = function (func) {
                var arr = [];
                for (var i = 0; i < this.length; i++)
                    if (isNil(func) || func(this[i], i) === true)
                        arr.push(this[i]);
                return new ArrayX(arr);
            };
            Arr["orderBy"] = function (func) {
                return new ArrayX(this
                    .select()
                    .sort(function (a, b) {
                    var objA = func ? func(a) : a;
                    var objB = func ? func(b) : b;
                    var comparison = null;
                    if (objA && objB) {
                        if (type(objA) === "number")
                            comparison = objA - objB;
                        else if (type(objA) === "date")
                            comparison = objA.getTime() - objB.getTime();
                        else
                            comparison = objA.toString().localeCompare(objB.toString());
                    }
                    else {
                        if (!objA && !objB)
                            comparison = 0;
                        else if (!objA)
                            comparison = -1;
                        else
                            comparison = 1;
                    }
                    return comparison;
                }));
            };
            Arr["orderByDesc"] = function (func) {
                return new ArrayX(this.orderBy(func).reverse());
            };
            Arr["firstOrDefault"] = function (predicate) {
                if (!predicate)
                    return this.length > 0 ? this[0] : null;
                for (var i = 0; i < this.length; i++)
                    if (predicate(this[i]) === true)
                        return this[i];
                return null;
            };
            Arr["first"] = function (predicate) {
                var firstOrDefault = this.firstOrDefault(predicate);
                if (isNil(firstOrDefault))
                    throw new Error("First item not found");
                else
                    return firstOrDefault;
            };
            Arr["lastOrDefault"] = function (predicate) {
                if (isNil(predicate))
                    return this.length > 0 ? this[this.length - 1] : null;
                for (var i = (this.length - 1); i >= 0; i--)
                    if (predicate(this[i]) === true)
                        return this[i];
                return null;
            };
            Arr["last"] = function (predicate) {
                var lastOrDefault = this.lastOrDefault(predicate);
                if (isNil(lastOrDefault))
                    throw new Error("Last item not found");
                else
                    return lastOrDefault;
            };
            Arr["all"] = function (predicate) {
                return this.where(predicate).length === this.length;
            };
            Arr["any"] = function (predicate) {
                return this.firstOrDefault(predicate) !== null;
            };
            Arr["take"] = function (count) {
                var arr = [];
                for (var i = 0; (i < count && i < this.length); i++)
                    arr.push(this[i]);
                return new ArrayX(arr);
            };
            Arr["skip"] = function (count) {
                var arr = [];
                for (var i = count; i < this.length; i++)
                    arr.push(this[i]);
                return new ArrayX(arr);
            };
            Arr["sum"] = function (func) {
                var sum = 0;
                for (var i = 0; i < this.length; i++)
                    sum += (func ? func(this[i]) : 0);
                return sum;
            };
            Arr["avg"] = function (func) {
                var count = this.count(func);
                if (count === 0)
                    return 0;
                else
                    return this.sum(func) / count;
            };
            Arr["count"] = function (func) {
                var count = 0;
                for (var i = 0; i < this.length; i++)
                    if (true === (func ? func(this[i]) : this[i]))
                        count++;
                return count;
            };
            Arr["min"] = function (func) {
                var min = Number.POSITIVE_INFINITY;
                for (var i = 0; i < this.length; i++) {
                    var value = (func ? func(this[i]) : this[i]);
                    if (value < min)
                        min = value;
                }
                return min;
            };
            Arr["max"] = function (func) {
                var max = Number.NEGATIVE_INFINITY;
                for (var i = 0; i < this.length; i++) {
                    var value = (func ? func(this[i]) : this[i]);
                    if (value > max)
                        max = value;
                }
                return max;
            };
            /*
             * Groups the array values into a dictionary of arrays.
             * Grouping keys are determined by applying the <func> function arg to each element
             */
            Arr["groupBy"] = function (func) {
                var groups = new associative.Dictionary();
                for (var i = 0; i < this.length; i++) {
                    var key = func(this[i]);
                    if (!groups.hasKey(key))
                        groups.put(key, []);
                    groups.get(key).push(this[i]);
                }
                return groups;
            };
            /*
             * Groups this array into an dictionary of <keyCount> arrays,
             * each containing the same number of elements, except possibly the last array,
             * which may have less elements than the others
             * if the length of this array is not divisible by <keyCount>
             */
            Arr["groupByNumber"] = function (keyCount) {
                var groups = new associative.Dictionary();
                var groupLength = Math.floor(this.length / keyCount);
                var remainder = this.length % this.count();
                for (var i = 0; i < keyCount; i++) {
                    if (!groups.hasKey(i))
                        groups.put(i, []);
                    groups.get(i).append(this.toArray().slice(i * groupLength, (i + 1) * groupLength));
                }
                return groups;
            };
            /*
             * groups this array according to <key> generator function and then
             * aggregates the resulting dictionary over <value> generator function
             */
            Arr["groupAggregate"] = function (key, value, aggregation) {
                // if key is passed as generator function, use it, else create generator function that assumes key is a property name. 
                // If key is null/missing, the string representation of an object is used as the key. 
                var keyGenerator = type(key) === "function"
                    ? key
                    : function (datum) { return key
                        ? datum[key.toString()]
                        : datum.toString(); };
                //Same for value
                var valueGenerator = type(value) === "function"
                    ? value
                    : function (datum) { return value
                        ? datum[value.toString()]
                        : datum.toString(); };
                return this
                    .groupBy(keyGenerator)
                    .select(function (groupData) {
                    return groupData[aggregation](valueGenerator);
                });
            };
            Arr["removeItem"] = function (item) {
                var arr = clone(this);
                var _index = arr.indexOf(item);
                while (_index >= 0) {
                    arr.splice(_index, 1);
                    _index = arr.indexOf(item);
                }
                return new ArrayX(arr);
            };
            Arr["removeAt"] = function (index, count) {
                var arr = clone(this);
                arr.splice(index, count);
                return new ArrayX(arr);
            };
            Arr["insert"] = function (index) {
                var items = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    items[_i - 1] = arguments[_i];
                }
                var arr = clone(this);
                arr.splice.apply(arr, [index, 0].concat(items));
                return new ArrayX(arr);
            };
            Arr["append"] = function (items) {
                var arr = clone(this);
                arr.splice.apply(arr, [this.length, 0].concat(items));
                return new ArrayX(arr);
            };
            Arr["flatten"] = function () {
                var mutable = false;
                var array = this;
                var toString = Object.prototype.toString;
                var arrayTypeStr = '[object Array]';
                var result = [];
                var nodes = (mutable && array) || array.slice();
                var node;
                if (!array.length) {
                    return new ArrayX(result);
                }
                node = nodes.pop();
                do {
                    if (toString.call(node) === arrayTypeStr) {
                        nodes.push.apply(nodes, node);
                    }
                    else {
                        result.push(node);
                    }
                } while (nodes.length && (node = nodes.pop()) !== undefined);
                result.reverse(); // we reverse result to restore the original order
                return new ArrayX(result);
            };
            /**
             * Determines whether an array includes a certain element, returning true or false as appropriate.
             * @param searchElement The element to search for.
             * @param fromIndex The position in this array at which to begin searching for searchElement.
             */
            Arr["contains"] = function (obj, fromIndex) {
                return this.indexOf(obj, fromIndex) >= 0;
            };
            return Arr;
        }
        ArrayX.fromLength = function (size) {
            if (size === void 0) { size = 0; }
            return new ArrayX(new ArrayX(size));
        };
        ArrayX.fromRange = function (from, to) {
            var _arr = [];
            var sgn = to > from ? 1 : -1;
            for (var idx = from; idx != to; idx += sgn)
                _arr.push(idx);
            return new ArrayX(_arr);
        };
        return ArrayX;
    }());
    ;
    return ArrayX;
})();
