declare namespace X {
    type Func = (currentValue: any, index?: number) => any
    type Predicate = (currentValue: any, index?: number) => boolean
    type Accumulator = (accumulatedValue: any, currentValue: any, index?: number) => any

    class Array<T> implements ArrayLike<T> {
        constructor(...items: T[]);

        toArray: () => any[];
        distinct: () => Array<T>;
        accumulate: (initial: any, accum?: Accumulator) => any;
        union: (array: Array<any>) => Array<T>;

        select: (func?: Func) => Array<T>;
        where: (func?: Predicate) => Array<T>;

        orderBy: (func?: Func) => Array<T>;
        orderByDesc: (func: Func) => Array<T>;

        firstOrDefault: (predicate?: Predicate) => T;
        first: (predicate?: Predicate) => T;

        lastOrDefault: (predicate?: Predicate) => T;
        last: (predicate?: Predicate) => T;

        all: (predicate?: Predicate) => boolean;
        any: (predicate?: Predicate) => boolean;

        take: (count: number) => Array<T>;
        skip: (count: number) => Array<T>;

        sum: (func?: Func) => number;
        avg: (func?: Func) => number;
        count: (func?: Func) => number;
        min: (func?: Func) => number;
        max: (func?: Func) => number;

        /* 
         * Groups the array values into a dictionary of arrays. 
         * Grouping keys are determined by applying the <func> function arg to each element 
         */
        groupBy: (func) => any;

        /* 
         * Groups this array into an dictionary of <keyCount> arrays,
         * each containing the same number of elements, except possibly the last array, 
         * which may have less elements than the others 
         * if the length of this array is not divisible by <keyCount>
         */
        groupByNumber(keyCount: number): any;

        /*
         * groups this array according to <key> generator function and then 
         * aggregates the resulting dictionary over <value> generator function
         */
        groupAggregate(key, value, aggregation): any;

        removeItem: (item: any) => Array<T>;
        removeAt: (index: number, count?: number) => Array<T>;
        insert: (index: number, ...items: any[]) => Array<T>;
        append: (items: any[]) => Array<T>;

        flatten: () => Array<T>;

        /**
         * Determines whether an array includes a certain element, returning true or false as appropriate.
         * @param searchElement The element to search for.
         * @param fromIndex The position in this array at which to begin searching for searchElement.
         */
        contains(obj: T, fromIndex?: number): boolean;

        static fromSize: (size?: number) => Array<any>;
        static fromRange: (from: number, to: number) => Array<any>;
        
        /**
         * Gets the length of the array. This is a number one higher than the highest element defined in an array.
         */
        readonly length: number;
        readonly [n: number]: T;
    }

}